import json

import requests


def get_token():
    response = requests.post('https://danielvaughan.eu.auth0.com/oauth/token',
                      data='{"client_id":"Zdsog4nDAnhQ99yiKwMQWAPc2qUDlR99","client_secret":"t-OAE-GQk_nZZtWn-QQezJxDsLXmU7VSzlAh9cKW5vb87i90qlXGTvVNAjfT9weF","audience":"http://localhost:8080","grant_type":"client_credentials"}',
                      headers={'Content-type': 'application/json'})
    data = json.loads(response.text)
    return data['access_token']



#!/usr/bin/env python
"""
This script will read a spreadsheet, generate a manifest, submit all items to the ingest API, 
assign uuid and generate a directory of bundles for the submitted data
"""
from openpyxl.utils.exceptions import InvalidFileException

__author__ = "jupp"
__license__ = "Apache 2.0"

import json, os
import logging
import ingest.utils.token_util as token_util


from openpyxl import load_workbook
from ingest.api.ingestapi import IngestApi

from optparse import OptionParser

from itertools import chain
from collections import defaultdict


# these are spreadsheet fields that can be a list
# todo - these should be read out of the json schema at the start

schema_arrayFields = {
    "cell_line": ["genus_species", "publications", "ncbi_taxon_id", "supplementary_files", "process_ids"],
    "cell_suspension": ["genus_species", "target_cell_type", "ncbi_taxon_id", "supplementary_files", "process_ids"],
    "donor_organism": ["genus_species", "disease", "familial_relationship", "ethnicity", "strain", "ncbi_taxon_id",
                       "supplementary_files", "process_ids"],
    "organoid": ["genus_species", "ncbi_taxon_id", "supplementary_files", "process_ids"],
    "specimen_from_organism": ["genus_species", "disease", "ncbi_taxon_id", "supplementary_files", "process_ids",
                               "gross_image", "microscopic_image"],
    "sequence_file": ["insdc_run"],
    "analysis_process": ["inputs", "tasks", "input_bundles", "outputs", "operator_identity", "protocol_ids"],
    "collection_process": ["process_reagents", "operator_identity", "protocol_ids"],
    "dissociation_process": ["process_reagents", "operator_identity", "protocol_ids"],
    "imaging_process": ["field_counts", "field_microns", "field_resolution", "operator_identity", "protocol_ids"],
    "enrichment_process": ["process_reagents", "operator_identity", "protocol_ids"],
    "library_preparation_process": ["process_reagents", "operator_identity", "protocol_ids"],
    "sequencing_process": ["process_reagents", "operator_identity", "protocol_ids"],
    "project": ["contributors", "supplementary_files", "publications", "authors"],
    "publication": ["authors"]
}

schema_timeFields = {
    "cell_line": ["date_established"],
    "donor_organism": ["time_of_death"],
    "specimen_from_organism": ["collection_time"],
    "analysis_process": ["start_time", "stop_time", "timestamp_start_utc", "timestamp_stop_utc"],
    "collection_process": ["start_time"],
    "dissociation_process": ["start_time"],
    "enrichment_process": ["start_time"],
    "imaging_process": ["start_time"],
    "library_preparation_process": ["start_time"],
    "sequencing_process": ["start_time"]
}

schema_booleanFields = {
    "donor_organism": ["is_living", "cold_perfused"],
    "sequencing_process": ["paired_ends"],
    "death": ["cold_perfused"]
}

schema_integerFields = {
    "cell_line": ["ncbi_taxon_id", "passage_number"],
    "cell_suspension": ["total_estimated_cells", "ncbi_taxon_id", "passage_number"],
    "donor_organism": ["ncbi_taxon_id", "hardy_scale"],
    "organoid": ["ncbi_taxon_id", "passage_number"],
    "specimen_from_organism": ["ncbi_taxon_id", "ischemic_time", "postmortem_interval"],
    "sequence_file": ["lane_index", "read_length"],
    "analysis_process": ["cpus"],
    "imaging_process": ["field_counts", "field_microns", "field_resolution"],
    "library_preparation_process": ["spike_in_dilution", "barcode_offset", "barcode_length"],
    "project": ["pmid"],
    "project.publications": ["pmid"]
}

schema_numberFields = {
    "cell_morphology": ["cell_size", "cell_viability", "percent_necrosis"],
    "imaging_process": ["exposure_time"],
    "enrichment_process": ["min_size_selected", "max_size_selected"],
    "organoid": ["organoid_age", "cell_size", "cell_viability", "percent_necrosis"],
    "cell_line": ["cell_size", "cell_viability", "percent_necrosis"],
    "cell_suspension": ["cell_size", "cell_viability", "percent_necrosis"],
    "donor_organism": ["days_on_ventilator", "body_mass_index"],
    "specimen_from_organism": ["storage_time"]
}

# maps sheets to the latest version of each schema
# todo - this should be replaced by dynamic lookups against ingest-core /schemas endpoint
# (spec for lookup: query by 'title' (sheet name in the spreadsheet, field in the schema)
schema_sheetname_mappings = {
    # biomaterials
    "cell_line": "https://schema.humancellatlas.org/type/biomaterial/5.1.0/cell_line",
    "cell_suspension": "https://schema.humancellatlas.org/type/biomaterial/5.1.0/cell_suspension",
    "donor_organism": "https://schema.humancellatlas.org/type/biomaterial/5.1.0/donor_organism",
    "organoid": "https://schema.humancellatlas.org/type/biomaterial/5.1.0/organoid",
    "specimen_from_organism": "https://schema.humancellatlas.org/type/biomaterial/5.1.0/specimen_from_organism",
    # files
    "analysis_file": "https://schema.humancellatlas.org/type/file/5.1.0/analysis_file",
    "sequence_file": "https://schema.humancellatlas.org/type/file/5.1.0/sequence_file",
    # analysis processes
    "analysis_process": "https://schema.humancellatlas.org/type/process/analysis/5.1.0/analysis_process",
    # biomaterial processes
    "process": "https://schema.humancellatlas.org/type/process/1.0.0/process",
    "collection_process": "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/collection_process",
    "dissociation_process": "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/dissociation_process",
    "enrichment_process": "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/enrichment_process",
    # imaging processes
    "imaging_process": "https://schema.humancellatlas.org/type/process/imaging/5.1.0/imaging_process",
    # sequencing processes
    "library_preparation_process": "https://schema.humancellatlas.org/type/process/sequencing/5.1.0/library_preparation_process",
    "sequencing_process": "https://schema.humancellatlas.org/type/process/sequencing/5.1.0/sequencing_process",
    # projects
    "project": "https://schema.humancellatlas.org/type/project/5.1.0/project",
    # protocols
    "protocol": "https://schema.humancellatlas.org/type/protocol/5.1.0/protocol",
    "analysis_protocol": "https://schema.humancellatlas.org/type/protocol/analysis/5.1.0/analysis_protocol",
    "biomaterial_collection_protocol": "https://schema.humancellatlas.org/type/protocol/biomaterial/5.1.0/biomaterial_collection_protocol",
    "imaging_protocol": "https://schema.humancellatlas.org/type/protocol/imaging/5.1.0/imaging_protocol",
    "sequencing_protocol": "https://schema.humancellatlas.org/type/protocol/sequencing/5.1.0/sequencing_protocol"
}


class SpreadsheetSubmission:

    def __init__(self, dry=False, output=None, schema_version=None):
        # todo - the logging code below doesn't work in python 3 - we should upgrade this
        formatter = logging.Formatter(
            '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s')
        logging.basicConfig(formatter=formatter, level=logging.INFO)
        self.logger = logging.getLogger(__name__)

        self.dryrun = dry
        self.outputDir = output
        self.ingest_api = None
        if not self.dryrun:
            self.ingest_api = IngestApi()

    def createSubmission(self, token):
        self.logger.info("creating submission...")
        if not self.ingest_api:
            self.ingest_api = IngestApi()

        submissionUrl = self.ingest_api.createSubmission(token)
        self.logger.info("new submission " + submissionUrl)
        return submissionUrl

    # this only works for keys nested to two levels, need something smarter to handle arbitrary
    # depth keys e.g. we support <level1>.<level2> = <value>, where value is either a single value or list
    # this doesn't support <level1>.<level2>.<level3> except if <level3> is "ontology"
    #This function takes a dictionary object, a key/value pair from the spreadsheet input and an object type, and adds
    #the key/value pair to the dictionary. The type is used to assess whether a field is an ontology or an array based on the
    # two lookup dictionaries declared at the top.
    # For || separated strings, the value field is split. For json arrays, the value is put into an array. Json objects of type
    # ontology require some extra formatting, which is also done here.

    def _keyValueToNestedObject(self, obj, key, value, type):
        if "*" in key:
            key = key.replace("*", "")
        d = str(value)
        # If the value contains a double pipe (||) or the key is for a field that can be a list (with or without also being
        # an ontology field), put value into an array (splitting if necessary)
        if "\"" in str(value) or "||" in str(value) \
                or (type in schema_arrayFields.keys() and key.split('.')[-1] in schema_arrayFields[type]) \
                or (type in schema_arrayFields.keys() and (key.split('.')[-1] == "ontology" or key.split('.')[-1] == "text")
                    and key.split('.')[-2] in schema_arrayFields[type]):
            if "||" in str(value):
                d = str(value).split("||")
            else:
                d = [value]

            if key.split('.')[-1] in ["text", "ontology"]:

                if isinstance(d, list):
                    t = []
                    for index, v in enumerate(d):
                        t.append({key.split('.')[-1] : d[index]})
                    d = t
                else:
                    d = {key.split('.')[-1] : d}
                key = ".".join(key.split('.')[:-1])

        # Raise an error if the key is too nested
        if len(key.split('.')) > 3:
            raise ValueError('We don\'t support keys nested greater than 3 levels, found:'+key)

        # If the key is in the date_time field list, convert the date time into a string of format YYYY-MM-DDThh:mm:ssZ
        # so it validates. Same principle for integers and boolean types
        if type in schema_timeFields.keys():
            try:
                if key.split('.')[-1] in schema_timeFields[type]:
                    if isinstance(d, list):
                        for i, v in enumerate(d):
                            date_string = v.strftime("%Y-%m-%dT%H:%M:%SZ")
                            d[i] = date_string
                    else:
                        d = d.strftime("%Y-%m-%dT%H:%M:%SZ")
            except:
                self.logger.warn('Failed to convert field {0} in sheet {1} (value {2}) to date_time value'.format(key, type, d))
                d = str(d)
        if type in schema_integerFields.keys():
            try:
                if key.split('.')[-1] in schema_integerFields[type]:
                    if isinstance(d, list):
                        for i, v in enumerate(d):
                            d[i] = int(v)
                    else:
                        d = int(d)
            except:
                self.logger.warn('Failed to convert field {0} in sheet {1} (value {2}) to integer value'.format(key, type, d))
                d = str(d)
        if type in schema_numberFields.keys():
            try:
                if key.split('.')[-1] in schema_numberFields[type]:
                    if isinstance(d, list):
                        for i, v in enumerate(d):
                            d[i] = float(v)
                    else:
                        d = float(d)
            except:
                self.logger.warn(
                    'Failed to convert field {0} in sheet {1} (value {2}) to float value'.format(key, type, d))
                d = str(d)
        if type in schema_booleanFields.keys():
            try:
                if key.split('.')[-1] in schema_booleanFields[type]:
                    if isinstance(d, list):
                        for i, v in enumerate(d):
                            if v.lower() in ["true", "yes"]:
                                d[i] = True
                            elif v.lower() in ["false", "no"]:
                                d[i] = False
                    else:
                        if d.lower() in ["true", "yes"]:
                            d = True
                        elif d.lower() in ["false", "no"]:
                            d = False
            except:
                self.logger.warn('Failed to convert field {0} in sheet {1} (value {2}) to boolean value'.format(key, type, d))
                d = str(d)

        # Build up the key-value dictionary
        for part in reversed(key.split('.')):
            d = {part: d}

        return self._mergeDict(obj, d)

    def _mergeDict(self, dict1, dict2):
        dict3 = defaultdict(list)
        for k, v in chain(dict1.items(), dict2.items()):
            if k in dict3:
                if isinstance(v, dict):
                    dict3[k].update(self._mergeDict(dict3[k], v))
                elif isinstance(v, list) and isinstance(dict3[k], list) and len(v) == len(dict3[k]):
                    for index, e in enumerate(v):
                        dict3[k][index].update(self._mergeDict(dict3[k][index], e))
                else:
                    dict3[k].update(v)
            else:
                dict3[k] = v
        return dict3

    #sheets with one or more data rows and properties in row 1
    def _multiRowToObjectFromSheet(self, type, sheet):
        objs = []
        for row in sheet.iter_rows(row_offset=3, max_row=(sheet.max_row - 3)):
            obj = {}
            hasData = False
            for cell in row:
                # WARNING: remove long from the checks if using python 3!
                if not cell.value and not isinstance(cell.value, (int, float, long)):
                    continue
                hasData = True
                cellCol = cell.col_idx
                propertyValue = sheet.cell(row=3, column=cellCol).value

                d = self._keyValueToNestedObject(obj, propertyValue, cell.value, type)
                obj.update(d)
            if hasData:
                self.logger.debug(json.dumps(obj))
                objs.append(obj)

        return objs

    # sheets that represent one entity where the properties are in column 0
    def _sheetToObject(self, type, sheet):
        obj = {}
        for row in sheet.iter_rows():
            if len(row) > 1:
                propertyCell = row[0].value
                valueCell = row[1].value
                if valueCell:
                    obj = self._keyValueToNestedObject(obj, propertyCell, valueCell, type)
                    # obj.update(d)
        self.logger.debug(json.dumps(obj))
        return obj

    def _emptyProcessObject(self, type, empty_process_id):
        obj = {}
        process_core = {"process_id": str(type) + "_process_" + str(empty_process_id)}
        schema_type = "process"
        describedBy = schema_sheetname_mappings["process"]
        obj["process_core"] = process_core
        obj["schema_type"] = schema_type
        obj["describedBy"] = describedBy
        return obj

    def completeSubmission(self):
        self.ingest_api.finishSubmission()

    def submit(self, pathToSpreadsheet, submissionUrl, token=None, project_id=None):
        try:
            self._process(pathToSpreadsheet, submissionUrl, token, project_id)
        except Exception as e:
            self.logger.error("Error:"+str(e))
            self.logger.error("Exception occurred processing spreadsheet", exc_info=e)
            raise e

    def dumpJsonToFile(self, object, projectId, name):
        if self.outputDir:
            dir = os.path.abspath(self.outputDir)
            if not os.path.exists(dir):
                os.makedirs(dir)
            tmpFile = open(dir + "/" + str(projectId) + "_" + str(name) + ".json", "w")
            tmpFile.write(json.dumps(object, indent=4))
            tmpFile.close()

    def _process(self, pathToSpreadsheet, submissionUrl, token, existing_project_id):

        # parse the spreadsheet
        try:
            wb = load_workbook(filename=pathToSpreadsheet)
        except InvalidFileException:
            raise SpreadsheetUploadError(400, "The uploaded file is not a valid XLSX spreadsheet", "")
        # This code section deals with cases where the project has already been submitted
        # ASSUMPTION: now additional project information (publications, contributors etc) is added via
        # the spreadsheet if the project already exists

        projectSheet = wb.create_sheet()
        projectPubsSheet = wb.create_sheet()
        contributorSheet = wb.create_sheet()
        empty_wrapper_id = 1
        empty_sampling_id = 1

        if existing_project_id is None:
            projectSheet = wb.get_sheet_by_name("project")

            if "project.publications" in wb.sheetnames:
                projectPubsSheet = wb.get_sheet_by_name("project.publications")
            if "contact" in wb.sheetnames:
                contributorSheet = wb.get_sheet_by_name("contact")

        specimenSheet = wb.create_sheet()
        donorSheet = wb.create_sheet()
        cellSuspensionSheet = wb.create_sheet()
        familialRelationshipSheet = wb.create_sheet()

        if "specimen_from_organism" in wb.sheetnames:
            specimenSheet = wb.get_sheet_by_name("specimen_from_organism")
        if "donor_organism" in wb.sheetnames:
            donorSheet = wb.get_sheet_by_name("donor_organism")
        if "cell_suspension" in wb.sheetnames:
            cellSuspensionSheet = wb.get_sheet_by_name("cell_suspension")
        if "familial_relationship" in wb.sheetnames:
            familialRelationshipSheet = wb.get_sheet_by_name("familial_relationship")

        organoidSheet = wb.create_sheet()
        if "organoid" in wb.sheetnames:
            organoidSheet = wb.get_sheet_by_name("organoid")

        clSheet = wb.create_sheet()
        if "cell_line" in wb.sheetnames:
            clSheet = wb.get_sheet_by_name("cell_line")

        clPublicationSheet = wb.create_sheet()
        if "cell_line.publications" in wb.sheetnames:
            clPublicationSheet = wb.get_sheet_by_name("cell_line.publications")


        protocolSheet = wb.create_sheet()
        collectionSheet = wb.create_sheet()
        dissociationSheet = wb.create_sheet()
        enrichmentSheet = wb.create_sheet()
        libraryPrepSheet = wb.create_sheet()
        sequencingSheet = wb.create_sheet()
        reagentsSheet = wb.create_sheet()
        filesSheet = wb.create_sheet()

        if "protocol" in wb.sheetnames:
            protocolSheet = wb.get_sheet_by_name("protocol")
        if "enrichment_process" in wb.sheetnames:
            enrichmentSheet = wb.get_sheet_by_name("enrichment_process")
        if "collection_process" in wb.sheetnames:
            collectionSheet = wb.get_sheet_by_name("collection_process")
        if "dissociation_process" in wb.sheetnames:
            dissociationSheet = wb.get_sheet_by_name("dissociation_process")
        if "library_preparation_process" in wb.sheetnames:
            libraryPrepSheet = wb.get_sheet_by_name("library_preparation_process")
        if "sequencing_process" in wb.sheetnames:
            sequencingSheet = wb.get_sheet_by_name("sequencing_process")
        if "purchased_reagents" in wb.sheetnames:
            reagentsSheet = wb.get_sheet_by_name("purchased_reagents")
        if "sequence_file" in wb.sheetnames:
            filesSheet = wb.get_sheet_by_name("sequence_file")


        # convert data in sheets back into dict + add the schema_type/describedBy boilerplate
        project = self._multiRowToObjectFromSheet("project", projectSheet)
        if project:
            if len(project) == 1:
                project = project[0]
                project.update({"schema_type": "project",
                  "describedBy": schema_sheetname_mappings["project"]})

        enrichment = self._multiRowToObjectFromSheet("enrichment_process", enrichmentSheet)
        if enrichment:
            for e in enrichment:
                e.update({"schema_type": "process",
                          "describedBy": schema_sheetname_mappings["enrichment_process"]})

        collection = self._multiRowToObjectFromSheet("collection_process", collectionSheet)
        if collection:
            for c in collection:
                c.update({"schema_type": "process",
                     "describedBy": schema_sheetname_mappings["collection_process"]})

        dissociation = self._multiRowToObjectFromSheet("dissociation_process", dissociationSheet)
        if dissociation:
            for d in dissociation:
                d.update({"schema_type": "process",
                  "describedBy": schema_sheetname_mappings["dissociation_process"]})

        reagents = self._multiRowToObjectFromSheet("purchased_reagents", reagentsSheet)
        libraryPrep = self._multiRowToObjectFromSheet("library_preparation_process", libraryPrepSheet)
        if libraryPrep:
            for lp in libraryPrep:
                lp.update({"schema_type": "process",
                    "describedBy": schema_sheetname_mappings["library_preparation_process"]})
        sequencing = self._multiRowToObjectFromSheet("sequencing_process", sequencingSheet)
        if sequencing:
            for s in sequencing:
                s.update({"schema_type": "process",
                  "describedBy": schema_sheetname_mappings["sequencing_process"]})

        protocols = self._multiRowToObjectFromSheet("protocol", protocolSheet)
        if protocols:
            for prot in protocols:
                prot.update({"schema_type": "protocol",
                  "describedBy": schema_sheetname_mappings["protocol"]})
        donors = self._multiRowToObjectFromSheet("donor_organism", donorSheet)
        if donors:
            for do in donors:
                do.update({"schema_type": "biomaterial",
                  "describedBy": schema_sheetname_mappings["donor_organism"]})
        familialRelationships = self._multiRowToObjectFromSheet("familial_relationship", familialRelationshipSheet)
        publications = self._multiRowToObjectFromSheet("project.publications", projectPubsSheet)
        contributors = self._multiRowToObjectFromSheet("contributor", contributorSheet)

        specimens = self._multiRowToObjectFromSheet("specimen_from_organism", specimenSheet)
        if specimens:
            for spec in specimens:
                spec.update({"schema_type": "biomaterial",
                  "describedBy": schema_sheetname_mappings["specimen_from_organism"]})
        cell_suspension = self._multiRowToObjectFromSheet("cell_suspension", cellSuspensionSheet)
        if cell_suspension:
            for cs in cell_suspension:
                cs.update({"schema_type": "biomaterial",
                     "describedBy": schema_sheetname_mappings["cell_suspension"]})
        organoid = self._multiRowToObjectFromSheet("organoid", organoidSheet)
        if organoid:
            for org in organoid:
                org.update({"schema_type": "biomaterial",
                  "describedBy": schema_sheetname_mappings["organoid"]})
        cell_line = self._multiRowToObjectFromSheet("cell_line", clSheet)
        if cell_line:
            for cl in cell_line:
                cl.update({"schema_type": "biomaterial",
                  "describedBy": schema_sheetname_mappings["cell_line"]})
        cell_line_publications = self._multiRowToObjectFromSheet("cell_line.publications", clPublicationSheet)
        files = self._multiRowToObjectFromSheet("sequence_file", filesSheet)
        if files:
            for f in files:
                f.update({"schema_type": "file",
                 "describedBy": schema_sheetname_mappings["sequence_file"]})


        biomaterials = []
        biomaterials.extend(donors)
        biomaterials.extend(specimens)
        biomaterials.extend(cell_suspension)
        biomaterials.extend(organoid)
        biomaterials.extend(cell_line)

        processes = []
        processes.extend(collection)
        processes.extend(dissociation)
        processes.extend(enrichment)
        processes.extend(libraryPrep)
        processes.extend(sequencing)

        chained_processes = []

        # creating submission
        #
        if not self.dryrun and not submissionUrl:
            token = "Bearer " + token_util.get_token()
            submissionUrl = self.createSubmission(token)

        # post objects to the Ingest API after some basic validation
        if existing_project_id is None:
            self.logger.info("Creating a new project for the submission")
            if "project_shortname" not in project["project_core"]:
                raise ValueError('Project must have an id attribute')
            projectId = project["project_core"]["project_shortname"]

             # embedd contact & publication into into project for now
            pubs = []
            for index, publication in enumerate(publications):
                if "project_core" in publication and "project_shortname" in publication["project_core"] and publication["project_core"]["project_shortname"] == projectId:
                    del publication["project_core"]
                else:
                    raise ValueError('Publication must reference the correct project shortname')
                pubs.append(publication)
            project["publications"] = pubs

            cont = []
            for index, contributor in enumerate(contributors):
                if "project_core" in contributor and "project_shortname" in contributor["project_core"] and contributor["project_core"]["project_shortname"] == projectId:
                    del contributor["project_core"]
                else:
                    raise ValueError('Contributor must reference the correct project shortname')
                cont.append(contributor)
            project["contributors"] = cont

            self.dumpJsonToFile(project, projectId, "project")

            projectIngest = None
            if not self.dryrun:
                projectIngest = self.ingest_api.createProject(submissionUrl, json.dumps(project), token)

        else:
            if not self.dryrun:
                self.logger.info("Retrieving existing project: " + existing_project_id)
                projectIngest = self.ingest_api.getProjectById(existing_project_id)
                submissionEnvelope = self.ingest_api.getSubmissionEnvelope(submissionUrl)
                self.ingest_api.linkEntity(projectIngest, submissionEnvelope, "submissionEnvelopes") # correct
            else:
                projectIngest = {"content" :
                                     {"project_core":
                                     {"project_shortname" : "dummy_project_id"}}}

            projectId = projectIngest["content"]["project_core"]["project_shortname"]

            self.dumpJsonToFile(projectIngest, projectId, "existing_project")

        protocolMap = {}
        for index, protocol in enumerate(protocols):
            if "protocol_id" not in protocol["protocol_core"]:
                raise ValueError('Protocol must have an id attribute')
            self.dumpJsonToFile(protocol, projectId, "protocol_" + str(index))
            protocolMap[protocol["protocol_core"]["protocol_id"]] = protocol
            if not self.dryrun:
                protocolIngest = self.ingest_api.createProtocol(submissionUrl, json.dumps(protocol))
                # self.ingest_api.linkEntity(protocolIngest, projectIngest, "projects")
                protocolMap[protocol["protocol_core"]["protocol_id"]] = protocolIngest

        biomaterialMap = {}


        for index, biomaterial in enumerate(biomaterials):
            if "biomaterial_id" not in biomaterial["biomaterial_core"]:
                raise ValueError('Biomaterial must have an id attribute')
            biomaterialMap[biomaterial["biomaterial_core"]["biomaterial_id"]] = biomaterial

            if "ncbi_taxon_id" in biomaterial["biomaterial_core"] and "genus_species" in biomaterial:
                if not isinstance(biomaterial["genus_species"],list):
                    biomaterial["genus_species"]["ontology"] = "NCBITaxon:" + str(biomaterial["biomaterial_core"]["ncbi_taxon_id"])
                elif isinstance(biomaterial["genus_species"],list) and len(biomaterial["genus_species"])==1:
                    biomaterial["genus_species"][0]["ontology"] = "NCBITaxon:" + str(biomaterial["biomaterial_core"]["ncbi_taxon_id"][0])


        # add dependent information to various biomaterial types

        for publication in cell_line_publications:
            if "biomaterial_id" in publication["biomaterial_core"]:
                bio_id = publication["biomaterial_core"]["biomaterial_id"]
                del publication["biomaterial_core"]

                if "publications" not in biomaterialMap[bio_id]:
                    biomaterialMap[bio_id]["publications"] = []
                biomaterialMap[bio_id]["publications"].append(publication)

        for familialRel in familialRelationships:
            if "biomaterial_id" in familialRel["biomaterial_core"]:
                bio_id = familialRel["biomaterial_core"]["biomaterial_id"]
                del familialRel["biomaterial_core"]

                if "familial_relationship" not in biomaterialMap[bio_id]:
                    biomaterialMap[bio_id]["familial_relationship"] = []
                biomaterialMap[bio_id]["familial_relationship"].append(familialRel)

        # build the process map from the different types of process infromation
        processMap = {}
        chainedProcessMap = {}
        for index, process in enumerate(processes):
            if "process_id" not in process["process_core"]:
                raise ValueError('Process must have an id attribute')
            processMap[process["process_core"]["process_id"]] = process

        for reagent in reagents:
            if "process_id" in reagent["process_core"]:
                process_id = reagent["process_core"]["process_id"]
                del reagent["process_core"]

                if "process_reagents" not in processMap[process_id]:
                    processMap[process_id]["process_reagents"] = []
                processMap[process_id]["process_reagents"].append(reagents)

        # submit biomaterials to ingest and link to project and protocols
        biomaterials_with_procs = []
        proc_input_biomaterials = {}
        biomaterial_proc_inputs = {}
        proc_output_biomaterials = {}
        biomaterial_proc_outputs = {}
        procs_wrapped_by = {}
        for index, biomaterial_id in enumerate(biomaterialMap.keys()):
            biomaterial = biomaterialMap[biomaterial_id]
            if "has_input_biomaterial" in biomaterial["biomaterial_core"]:
                if biomaterial["biomaterial_core"]["has_input_biomaterial"] not in biomaterialMap.keys():
                    raise ValueError('Biomaterial '+ str(biomaterial_id) +' references another biomaterial '+ str(biomaterial["biomaterial_core"]["has_input_biomaterial"]) +' that isn\'t in the spraedsheet')

            if "process_ids" in biomaterial:
                biomaterials_with_procs.append(biomaterial_id)
                # do we have multiple chained protocols? If so, create or reuse a 'wrapper'
                wrapper_process = {}
                if len(biomaterial["process_ids"]) > 1:
                    process_ids_field = str(biomaterial["process_ids"])
                    if process_ids_field in procs_wrapped_by:
                        wrapper_process = procs_wrapped_by[process_ids_field]
                    else:
                        wrapper_process = self._emptyProcessObject("wrapper", empty_wrapper_id)
                        empty_wrapper_id += 1
                        procs_wrapped_by[process_ids_field] = wrapper_process
                        processMap[wrapper_process["process_core"]["process_id"]] = wrapper_process
                        processMap[wrapper_process["process_core"]["process_id"]]["chained_process_ids"] = []

                for process_id in biomaterial["process_ids"]:
                    if process_id not in processMap.keys():
                        raise ValueError(
                         'A biomaterial references a process ' + process_id + ' that isn\'t in the biomaterials worksheet')

                    if biomaterial_id not in biomaterial_proc_inputs:
                        biomaterial_proc_inputs[biomaterial_id] = []

                    # do we have a wrapper process?
                    if wrapper_process:
                        # link this process to the wrapper
                        wrapper_id = wrapper_process["process_core"]["process_id"]
                        if "biomaterial_ids" not in processMap[wrapper_id]:
                            processMap[wrapper_id]["biomaterial_ids"] = []

                        processMap[wrapper_id]["chained_process_ids"].append(process_id)
                        if process_id not in chainedProcessMap:
                            chainedProcessMap[process_id] = []
                        if wrapper_id not in chainedProcessMap[process_id]:
                            chainedProcessMap[process_id].append(wrapper_id)

                        # link input or output biomaterials to the wrapper (i.e. indirectly)
                        if biomaterial_id not in processMap[wrapper_id]["biomaterial_ids"]:
                            processMap[wrapper_id]["biomaterial_ids"].append(biomaterial_id)
                        if wrapper_id not in proc_input_biomaterials:
                            proc_input_biomaterials[wrapper_id] = []
                        if biomaterial_id not in proc_input_biomaterials[wrapper_id]:
                            proc_input_biomaterials[wrapper_id].append(biomaterial_id)
                        if wrapper_id not in biomaterial_proc_inputs[biomaterial_id]:
                            biomaterial_proc_inputs[biomaterial_id].append(wrapper_id)
                    else:
                        # link processes to input and output directly
                        if "biomaterial_ids" not in processMap[process_id]:
                            processMap[process_id]["biomaterial_ids"] = []

                        processMap[process_id]["biomaterial_ids"].append(biomaterial["biomaterial_core"]["biomaterial_id"])
                        if process_id not in proc_input_biomaterials:
                            proc_input_biomaterials[process_id] = []
                        if biomaterial_id not in proc_input_biomaterials[process_id]:
                            proc_input_biomaterials[process_id].append(biomaterial_id)
                        if process_id not in biomaterial_proc_inputs[biomaterial_id]:
                            biomaterial_proc_inputs[biomaterial_id].append(process_id)
                del biomaterial["process_ids"]

            self.dumpJsonToFile(biomaterial, projectId, "biomaterial_" + str(index))
            if not self.dryrun:
                biomaterialIngest = self.ingest_api.createBiomaterial(submissionUrl, json.dumps(biomaterial))
                self.ingest_api.linkEntity(biomaterialIngest, projectIngest, "projects") # correct
                biomaterialMap[biomaterial["biomaterial_core"]["biomaterial_id"]] = biomaterialIngest

        # create has_input_biomaterial links between biomaterials separately to make sure all biomaterials are submitted
        for index, biomaterial_id in enumerate(biomaterialMap.keys()):
            if not self.dryrun:
                if "has_input_biomaterial" in biomaterialMap[biomaterial_id]['content']["biomaterial_core"]:
                    # retrieve biomaterials from map
                    output_biomaterial = biomaterialMap[biomaterial_id]
                    input_biomaterial = biomaterialMap[biomaterialMap[biomaterial_id]['content']["biomaterial_core"]["has_input_biomaterial"]]

                    # if the input biomaterial declares the process, we will link later
                    if input_biomaterial['content']['biomaterial_core']['biomaterial_id'] not in biomaterials_with_procs:
                        # else create sampling process to link biomaterials
                        sampling_process = self._emptyProcessObject("sampling", empty_sampling_id)
                        empty_sampling_id += 1
                        sampling_process_ingest = self.ingest_api.createProcess(submissionUrl, json.dumps(sampling_process))

                        # link process to input biomaterials
                        self.ingest_api.linkEntity(input_biomaterial, sampling_process_ingest, "inputToProcesses")
                        # link process to output biomaterials
                        self.ingest_api.linkEntity(output_biomaterial, sampling_process_ingest, "derivedByProcesses")
                    else:
                        for process_id in biomaterial_proc_inputs[input_biomaterial['content']['biomaterial_core']['biomaterial_id']]:
                            if process_id not in proc_output_biomaterials:
                                proc_output_biomaterials[process_id] = []
                            proc_output_biomaterials[process_id].append(biomaterial_id)
                            if biomaterial_id not in biomaterial_proc_outputs:
                                biomaterial_proc_outputs[biomaterial_id] = []
                            biomaterial_proc_inputs[biomaterial_id].append(process_id)

        filesMap={}
        for index, file in enumerate(files):
            if "file_name" not in file["file_core"]:
                raise ValueError('Files must have a name')
            if "process_id" not in file:
                raise ValueError('Files must be linked to a process')
            process = file["process_id"]
            del file["process_id"]
            del file["biomaterial_id"]
            filesMap[file["file_core"]["file_name"]] = file

            # is the process referred to from this file wrapped by another process?
            if process in chainedProcessMap:
                for wrapper_process in chainedProcessMap[process]:
                    if "files" not in processMap[wrapper_process]:
                        processMap[wrapper_process]["files"] = []
                    processMap[wrapper_process]["files"].append(file["file_core"]["file_name"])
            else:
                if "files" not in processMap[process]:
                    processMap[process]["files"] = []
                processMap[process]["files"].append(file["file_core"]["file_name"])

            self.dumpJsonToFile(file, projectId, "files_" + str(index))
            if not self.dryrun:
                fileIngest = self.ingest_api.createFile(submissionUrl, file["file_core"]["file_name"], json.dumps(file))
                filesMap[file["file_core"]["file_name"]] = fileIngest

        # create all the chained processes first, these will be referred to by wrapper processes
        chained_process_ingest_map = {}
        for index, chained_process in enumerate(chainedProcessMap.keys()):
            if chained_process not in processMap:
                raise ValueError('A chained process was not found in the process sheet - ' + str(chained_process))
            if not self.dryrun:
                chained_process_protocols = []
                if "protocol_ids" in processMap[chained_process]:
                    for protocol_id in processMap[chained_process]["protocol_ids"]:
                        if protocol_id not in protocolMap:
                            raise ValueError('An process references a protocol '+protocol_id+' that isn\'t in one of the protocol worksheets')
                        chained_process_protocols.append(protocolMap[protocol_id])
                    del processMap[chained_process]["protocol_ids"]

                chained_process_ingest = self.ingest_api.createProcess(submissionUrl, json.dumps(processMap[chained_process]))
                chained_process_ingest_map[chained_process] = chained_process_ingest
                for protocol in chained_process_protocols:
                    self.ingest_api.linkEntity(chained_process_ingest, protocol, "protocols")

        for index, process in enumerate(processMap.values()):
            if "process_id" not in process["process_core"]:
                raise ValueError('Each process must have an id attribute' + str(process))

            output_files=[]
            if "files" in process:
                for file in process["files"]:
                    if file not in filesMap:
                        raise ValueError('Process references file '+file+' that isn\'t defined in the files sheet')
                output_files = process["files"]
                del process["files"]

            output_biomaterials = []
            if "biomaterial_ids" not in process:
                # this is allowed if this is a chained process
                if process["process_core"]["process_id"] not in chainedProcessMap:
                    raise ValueError("Every process must reference a biomaterial using the biomaterial_id attribute")
            else:
                for biomaterial_id in process["biomaterial_ids"]:
                    if biomaterial_id not in biomaterialMap:
                        raise ValueError('An process references a biomaterial '+biomaterial_id+' that isn\'t in one of the biomaterials worksheets')
                output_biomaterials = process["biomaterial_ids"]
                del process["biomaterial_ids"]

            chained_processes=[]
            if "chained_process_ids" in process:
                for chained_process_id in process["chained_process_ids"]:
                    if not self.dryrun:
                        chained_processes.append(chained_process_ingest_map[chained_process_id])
                del process["chained_process_ids"]

            process_protocols = []
            if "protocol_ids" in process:
                for protocol_id in process["protocol_ids"]:
                    if protocol_id not in protocolMap:
                        raise ValueError('An process references a protocol '+protocol_id+' that isn\'t in one of the protocol worksheets')
                process_protocols = process["protocol_ids"]
                del process["protocol_ids"]

            self.dumpJsonToFile(process, projectId, "process_" + str(index))
            if not self.dryrun:

                process_id = process["process_core"]["process_id"]

                if process_id in chained_process_ingest_map:
                    processIngest = chained_process_ingest_map[process_id]
                else:
                    processIngest = self.ingest_api.createProcess(submissionUrl, json.dumps(process))

                self.ingest_api.linkEntity(processIngest, projectIngest, "projects") # correct

                if process["process_core"]["process_id"] in proc_input_biomaterials:
                    for biomaterial_id in proc_input_biomaterials[process["process_core"]["process_id"]]:
                        self.ingest_api.linkEntity(biomaterialMap[biomaterial_id], processIngest, "inputToProcesses")

                if process["process_core"]["process_id"] in proc_output_biomaterials:
                    for biomaterial_id in proc_output_biomaterials[process["process_core"]["process_id"]]:
                        self.ingest_api.linkEntity(biomaterialMap[biomaterial_id], processIngest, "derivedByProcesses")

                for file in output_files:
                    self.ingest_api.linkEntity(filesMap[file], processIngest, "derivedByProcesses") # correct

                for chained_process in chained_processes:
                    self.ingest_api.linkEntity(processIngest, chained_process, "chainedProcesses")  # correct

                for protocol_id in process_protocols:
                    self.ingest_api.linkEntity(processIngest, protocolMap[protocol_id], "protocols")

        self.logger.info("All done!")
        wb.close()
        return submissionUrl

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-p", "--path", dest="path",
                      help="path to HCA example data bundles", metavar="FILE")
    parser.add_option("-d", "--dry", help="doa dry run without submitting to ingest", action="store_true", default=False)
    parser.add_option("-o", "--output", dest="output",
                      help="output directory where to dump json files submitted to ingest", metavar="FILE", default=None)
    parser.add_option("-i", "--id", dest="project_id",
                      help="The project_id for an existing submission", default=None)
    parser.add_option("-v", "--version", dest="schema_version", help="Metadata schema version", default=None)


    (options, args) = parser.parse_args()
    if not options.path:
        print ("You must supply path to the HCA bundles directory")
        exit(2)
    submission = SpreadsheetSubmission(options.dry, options.output, options.schema_version)
    submission.submit(options.path, None, None, options.project_id)



#!/usr/bin/env python
"""
desc goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"


import json
import logging
import os

import sys
import uuid
from optparse import OptionParser

import ingest.api.stagingapi as stagingapi
import ingest.api.ingestapi as ingestapi
import ingest.api.dssapi as dssapi

import ingest.utils.bundlevalidator as bundlevalidator

DEFAULT_INGEST_URL = os.environ.get('INGEST_API', 'http://api.ingest.dev.data.humancellatlas.org')
DEFAULT_STAGING_URL = os.environ.get('STAGING_API', 'http://staging.dev.data.humancellatlas.org')
DEFAULT_DSS_URL = os.environ.get('DSS_API', 'http://dss.dev.data.humancellatlas.org')

BUNDLE_SCHEMA_BASE_URL = os.environ.get('BUNDLE_SCHEMA_BASE_URL', 'https://schema.humancellatlas.org/bundle/%s/')
METADATA_SCHEMA_VERSION = os.environ.get('SCHEMA_VERSION', '5.1.0')


class IngestExporter:
    def __init__(self, options=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        self.logger = logging.getLogger(__name__)

        self.dryrun = options.dry if options and options.dry else False
        self.outputDir = options.output if options and options.output else None

        self.ingestUrl = options.ingest if options and options.ingest else os.path.expandvars(DEFAULT_INGEST_URL)

        self.stagingUrl = options.staging if options and options.staging else os.path.expandvars(DEFAULT_STAGING_URL)
        self.dssUrl = options.dss if options and options.dss else os.path.expandvars(DEFAULT_DSS_URL)
        self.schema_version = options.schema_version if options and options.schema_version else os.path.expandvars(METADATA_SCHEMA_VERSION)
        self.schema_url = os.path.expandvars(BUNDLE_SCHEMA_BASE_URL % self.schema_version)

        self.logger.debug("ingest url is " + self.ingestUrl)

        self.staging_api = stagingapi.StagingApi()
        self.dss_api = dssapi.DssApi()
        self.ingest_api = ingestapi.IngestApi(self.ingestUrl)
        self.bundle_validator = bundlevalidator.BundleValidator()

    def writeBundleToFile(self, name, index, type, doc):
        dir = os.path.abspath("bundles/" + name)
        if not os.path.exists(dir):
            os.makedirs(dir)
        bundleDir = os.path.abspath(dir + "/bundle" + index)
        if not os.path.exists(bundleDir):
            os.makedirs(bundleDir)
        tmpFile = open(bundleDir + "/" + type + ".json", "w")
        tmpFile.write(json.dumps(self.bundleProject(doc), indent=4))
        tmpFile.close()

    def getSchemaNameForEntity(self, schemaUrl):
        return schemaUrl["content"]["describedBy"].rsplit('/', 1)[-1]

    def getLinks(self, source_type, source_id, destination_type, destination_id):
        return {
            'source_type': source_type,
            'source_id': source_id,
            'destination_type': destination_type,
            'destination_id': destination_id
        }

    def generateBundle(self, message):
        success = False
        callbackLink = message["callbackLink"]

        self.logger.info('process received ' + callbackLink)
        self.logger.info('process index: ' + str(message["index"]) + ', total processes: ' + str(message["total"]))

        # given an assay, generate a bundle

        processUrl = self.ingest_api.getAssayUrl(callbackLink)  # TODO rename getAssayUrl
        processUuid = message["documentUuid"]
        envelopeUuid = message["envelopeUuid"]

        # check staging area is available
        if self.dryrun or self.staging_api.hasStagingArea(envelopeUuid):
            assay = self.ingest_api.getAssay(processUrl)

            self.logger.info("Attempting to export bundle to DSS...")
            success = self.export_bundle(envelopeUuid, processUrl)
        else:
            error_message = "Can't do export as no upload area has been created"
            raise NoUploadAreaFoundError(error_message)

        if not success:
            raise Error("An error occurred in export. Failed to export to dss: " + message["callbackLink"])


    def bundleFileIngest(self, file_entity):
        return self._bundleEntityIngest(file_entity)

    def bundleProtocolIngest(self, protocol_entity):
        return self._bundleEntityIngest(protocol_entity)

    def _bundleEntityIngest(self, entity):
        return {
            'content': entity['content'],
            'hca_ingest': {
                'document_id': entity['uuid']['uuid'],
                'submissionDate': entity['submissionDate']
            }
        }

    def writeMetadataToStaging(self, submissionId, fileName, content, contentType):
        self.logger.info("writing to staging area..." + fileName)
        fileDescription = self.staging_api.stageFile(submissionId, fileName, content, contentType)
        self.logger.info("File staged at " + fileDescription.url)
        return fileDescription

    def bundleSample(self, sample_entity):
        sample_copy = self._copyAndTrim(sample_entity)
        bundle = {
            'content': sample_copy.pop('content', None),
            'hca_ingest': sample_copy
        }

        bundle["hca_ingest"]["document_id"] = bundle["hca_ingest"]["uuid"]["uuid"]
        del bundle["hca_ingest"]["uuid"]

        if bundle["hca_ingest"]["accession"] is None:
            bundle["hca_ingest"]["accession"] = ""
        return bundle

    def bundleProcess(self, process_entity):
        process_copy = self._copyAndTrim(process_entity)
        bundle = {
            'content': process_copy.pop('content', None),
            'hca_ingest': process_copy
        }

        bundle["hca_ingest"]["document_id"] = bundle["hca_ingest"]["uuid"]["uuid"]
        del bundle["hca_ingest"]["uuid"]

        if bundle["hca_ingest"]["accession"] is None:
            bundle["hca_ingest"]["accession"] = ""
        return bundle

    def bundleProject(self, project_entity):
        project_copy = self._copyAndTrim(project_entity)
        bundle = {
            'describedBy': "https://schema.humancellatlas.org/bundle/5.1.0/project",
            'schema_version': "5.1.0",
            'schema_type': 'project_bundle',
            'content': project_copy.pop('content', None),
            'hca_ingest': project_copy
        }

        bundle["hca_ingest"]["document_id"] = bundle["hca_ingest"]["uuid"]["uuid"]

        del bundle["hca_ingest"]["uuid"]

        if bundle["hca_ingest"]["accession"] is None:
            bundle["hca_ingest"]["accession"] = ""
        return bundle

    def bundleProtocol(self, protocol_entity):
        protocol_copy = self._copyAndTrim(protocol_entity)
        bundle = {
            'content': protocol_copy.pop('content', None),
            'hca_ingest': protocol_copy
        }

        bundle["hca_ingest"]["document_id"] = bundle["hca_ingest"]["uuid"]["uuid"]
        del bundle["hca_ingest"]["uuid"]

        if bundle["hca_ingest"]["accession"] is None:
            bundle["hca_ingest"]["accession"] = ""
        return bundle

    def _copyAndTrim(self, project_entity):
        copy = project_entity.copy()
        for property in ["_links", "events", "validationState", "validationErrors", "user", "lastModifiedUser"]:
            if property in copy:
                del copy[property]
        return copy

    def dumpJsonToFile(self, object, projectId, name, output_dir=None):
        if output_dir:
            self.outputDir = output_dir

        if self.outputDir:
            dir = os.path.abspath(self.outputDir)
            if not os.path.exists(dir):
                os.makedirs(dir)
            tmpFile = open(dir + "/" + projectId + "_" + name + ".json", "w")
            tmpFile.write(json.dumps(object, indent=4))
            tmpFile.close()

    def export_bundle(self, submission_uuid, process_url):
        self.logger.info('Export bundle for process: ' + process_url)

        saved_bundle_uuid = None

        self.logger.info('Retrieving all process information...')
        process_info = self.get_all_process_info(process_url)

        self.logger.info('Generating bundle files...')
        metadata_files_info = self.prepare_metadata_files(process_info)
        data_files_info = self.prepare_data_files(process_info)

        bundle_manifest = self.create_bundle_manifest(submission_uuid, metadata_files_info, process_info)

        if self.dryrun:
            self.logger.info('Export is using dry run mode.')
            self.logger.info('Dumping bundle files...')
            self.dump_metadata_files_and_bundle_manifest(metadata_files_info, bundle_manifest)

        else:
            self.logger.info('Uploading metadata files...')
            self.upload_metadata_files(submission_uuid, metadata_files_info)

            self.logger.info('Saving files in DSS...')
            bundle_uuid = bundle_manifest.bundleUuid

            metadata_files = self.get_metadata_files(metadata_files_info)
            data_files = self.get_data_files(data_files_info)
            bundle_files = metadata_files + data_files

            created_files = self.put_files_in_dss(bundle_uuid, bundle_files)

            self.logger.info('Saving bundle in DSS...')
            self.put_bundle_in_dss(bundle_uuid, created_files)

            self.logger.info('Saving bundle manifest...')
            self.ingest_api.createBundleManifest(bundle_manifest)

            saved_bundle_uuid = bundle_manifest.bundleUuid

            self.logger.info('Bundle ' + saved_bundle_uuid + ' was successfully created!')

        return saved_bundle_uuid

    def get_all_process_info(self, process_url):
        process = self.ingest_api.getAssay(process_url)  # TODO rename getAssay to getProcess
        process_info = ProcessInfo()
        process_info.input_bundle = self.get_input_bundle(process)

        process_info.project = self.get_project_info(process)

        if not process_info.project:  # get from input bundle
            project_uuid_lists = process_info.input_bundle['fileProjectMap'].values()

            if len(project_uuid_lists) == 0 and len(project_uuid_lists[0]) == 0:
                raise Error('Input bundle manifest has no list of project uuid.')  # very unlikely to happen

            project_uuid = project_uuid_lists[0][0]
            process_info.project = self.ingest_api.getProjectByUuid(project_uuid)

        self.recurse_process(process, process_info)

        return process_info

    def get_project_info(self, process):
        projects = list(self.ingest_api.getRelatedEntities('projects', process, 'projects'))

        if len(projects) > 1:
            raise MultipleProjectsError('Can only be one project in bundle')

        # TODO add checking for project only on an assay process
        # TODO an analysis process may have no link to a project

        if len(projects) > 0:
            return projects[0]

        return None

    # get all related info of a process
    def recurse_process(self, process, process_info):
        chained_processes = list(self.ingest_api.getRelatedEntities('chainedProcesses', process, 'processes'))

        is_wrapper = len(chained_processes) > 0

        # don't include wrapper processes in process bundle
        if is_wrapper:
            for chained_process in chained_processes:
                uuid = chained_process['uuid']['uuid']
                process_info.derived_by_processes[uuid] = chained_process
        else:
            uuid = process['uuid']['uuid']
            process_info.derived_by_processes[uuid] = process

        # get all derived by processes using input biomaterial and input files
        derived_by_processes = []

        # wrapper process has the links to input biomaterials and derived files to check if a process is an assay
        input_biomaterials = list(self.ingest_api.getRelatedEntities('inputBiomaterials', process, 'biomaterials'))
        for input_biomaterial in input_biomaterials:
            uuid = input_biomaterial['uuid']['uuid']
            process_info.input_biomaterials[uuid] = input_biomaterial
            derived_by_processes.extend(
                self.ingest_api.getRelatedEntities('derivedByProcesses', input_biomaterial, 'processes'))

        input_files = list(self.ingest_api.getRelatedEntities('inputFiles', process, 'files'))
        for input_file in input_files:
            uuid = input_file['uuid']['uuid']
            process_info.input_files[uuid] = input_file
            derived_by_processes.extend(
                self.ingest_api.getRelatedEntities('derivedByProcesses', input_file, 'processes'))

        derived_biomaterials = list(self.ingest_api.getRelatedEntities('derivedBiomaterials', process, 'biomaterials'))
        derived_files = list(self.ingest_api.getRelatedEntities('derivedFiles', process, 'files'))

        # since wrapper processes are not included in process bundle,
        #  links to it must be applied to its chained processes
        processes_to_link = chained_processes if is_wrapper else [process]
        for process_to_link in processes_to_link:
            process_name = self.getSchemaNameForEntity(process_to_link)
            process_uuid = process_to_link['uuid']['uuid']

            for input_biomaterial in input_biomaterials:
                uuid = input_biomaterial['uuid']['uuid']
                process_info.links.append(self.getLinks('biomaterial', uuid, process_name, process_uuid))

            for input_file in input_files:
                uuid = input_file['uuid']['uuid']
                process_info.links.append(self.getLinks('file', uuid, process_name, process_uuid))

            protocols = list(self.ingest_api.getRelatedEntities('protocols', process_to_link, 'protocols'))
            for protocol in protocols:
                uuid = protocol['uuid']['uuid']
                process_info.links.append(self.getLinks(process_name, process_uuid, 'protocol', uuid))
                process_info.protocols[uuid] = protocol

            for derived_file in derived_files:
                uuid = derived_file['uuid']['uuid']
                process_info.links.append(self.getLinks(process_name, process_uuid, 'file', uuid))
                process_info.derived_files[uuid] = derived_file

            for derived_biomaterial in derived_biomaterials:
                uuid = derived_biomaterial['uuid']['uuid']
                process_info.links.append(self.getLinks(process_name, process_uuid, 'biomaterial', uuid))

        for derived_by_process in derived_by_processes:
            self.recurse_process(derived_by_process, process_info)

    def get_input_bundle(self, process):
        bundle_manifests = list(self.ingest_api.getRelatedEntities('inputBundleManifests', process, 'bundleManifests'))

        if len(bundle_manifests) > 0:
            return bundle_manifests[0]

        return None

    def prepare_metadata_files(self, process_info):
        bundle_content = self.build_and_validate_content(process_info)

        metadata_files = {}

        file_uuid = str(uuid.uuid4())
        metadata_files['project'] = {
            'content': bundle_content['project'],
            'content_type': '"metadata/project"',
            'indexed': True,
            'dss_filename': 'project.json',
            'dss_uuid': file_uuid,
            'upload_filename': 'project_bundle_' + file_uuid + '.json'
        }

        file_uuid = str(uuid.uuid4())
        metadata_files['biomaterial'] = {
            'content': bundle_content['biomaterial'],
            'content_type': '"metadata/biomaterial"',
            'indexed': True,
            'dss_filename': 'biomaterial.json',
            'dss_uuid': file_uuid,
            'upload_filename': 'biomaterial_bundle_' + file_uuid + '.json'
        }

        file_uuid = str(uuid.uuid4())
        metadata_files['process'] = {
            'content': bundle_content['process'],
            'content_type': '"metadata/process"',
            'indexed': True,
            'dss_filename': 'process.json',
            'dss_uuid': file_uuid,
            'upload_filename': 'process_bundle_' + file_uuid + '.json'
        }

        file_uuid = str(uuid.uuid4())
        metadata_files['protocol'] = {
            'content': bundle_content['protocol'],
            'content_type': '"metadata/protocol"',
            'indexed': True,
            'dss_filename': 'protocol.json',
            'dss_uuid': file_uuid,
            'upload_filename': 'protocol_bundle_' + file_uuid + '.json'
        }

        file_uuid = str(uuid.uuid4())
        metadata_files['file'] = {
             'content': bundle_content['file'],
             'content_type': '"metadata/file"',
             'indexed': True,
             'dss_filename': 'file.json',
             'dss_uuid': file_uuid,
             'upload_filename': 'file_bundle_' + file_uuid + '.json'
        }

        file_uuid = str(uuid.uuid4())
        metadata_files['links'] = {
            'content': bundle_content['links'],
            'content_type': '"metadata/links"',
            'indexed': True,
            'dss_filename': 'links.json',
            'dss_uuid': file_uuid,
            'upload_filename': 'links_bundle_' + file_uuid + '.json'
        }

        self._inherit_same_files_from_input(metadata_files, process_info)

        return metadata_files

    # if new file has same set of uuids as the input bundle file
    # do not re-upload or create bundle metadata file in dss
    # reuse the file uuid
    # this scenario might only happen for project, biomaterial, protocol map
    def _inherit_same_files_from_input(self, metadata_files, process_info):
        input_bundle = process_info.input_bundle

        if not input_bundle:
            return

        file_uuids = [process_info.project['uuid']['uuid']]
        input_file = self._compare_to_input_file(input_bundle, 'fileProjectMap', file_uuids)
        if input_file['is_equal']:
            metadata_files['project']['dss_uuid'] = input_file['file_uuid']
            metadata_files['project']['is_same_as_input'] = input_file['is_equal']

        file_uuids = process_info.input_biomaterials.keys()
        input_file = self._compare_to_input_file(input_bundle, 'fileBiomaterialMap', file_uuids)
        if input_file['is_equal']:
            metadata_files['biomaterial']['dss_uuid'] = input_file['file_uuid']
            metadata_files['biomaterial']['is_same_as_input'] = input_file['is_equal']

        file_uuids = process_info.protocols.keys()
        input_file = self._compare_to_input_file(input_bundle, 'fileProtocolMap', file_uuids)
        if input_file['is_equal']:
            metadata_files['protocol']['dss_uuid'] = input_file['file_uuid']
            metadata_files['protocol']['is_same_as_input'] = input_file['is_equal']

    def _compare_to_input_file(self, input_bundle, attr, file_uuids):
        file_map = input_bundle[attr]
        input_file_uuids = list(file_map.values())[0]
        input_file_uuid = list(file_map.keys())[0]

        is_same_as_input_file = self._are_equal_lists(input_file_uuids, file_uuids)

        return {
            'is_equal': is_same_as_input_file,
            'file_uuid': input_file_uuid
        }

    # compare two lists ignoring order
    def _are_equal_lists(self, list1, list2):
        set1 = frozenset(list(list1))
        set2 = frozenset(list(list2))
        diff = set1.difference(set2)

        return not len(diff)

    # build bundle json for each entity according to schema
    def build_and_validate_content(self, process_info):
        bundle_contents = {
            'project': self.bundleProject(process_info.project),
            'biomaterial': self.bundle_biomaterials(process_info.input_biomaterials.values()),
            'process': self.bundle_processes(process_info.derived_by_processes.values()),
            'file': self.bundle_files(list(process_info.input_files.values()) + list(process_info.derived_files.values())),
            'protocol': self.bundle_protocols(process_info.protocols.values()),
            'links': self.bundle_links(process_info.links)
        }

        # TODO comment out for now, validation seems broken
        # self.validate_metadata_files(bundle_contents)

        return bundle_contents

    def bundle_biomaterials(self, biomaterials):
        return {
            'describedBy': 'https://schema.humancellatlas.org/bundle/5.1.0/biomaterial',
            'schema_version': '5.1.0',
            'schema_type': 'biomaterial_bundle',
            'biomaterials': list(map(self.bundleSample, biomaterials))
        }

    def bundle_processes(self, processes):
        return {
            'describedBy': 'https://schema.humancellatlas.org/bundle/5.2.1/process',
            'schema_version': '5.2.1',
            'schema_type': 'process_bundle',
            'processes': list(map(self.bundleProcess, processes))
        }

    def bundle_files(self, files):
        return {
            'describedBy': 'https://schema.humancellatlas.org/bundle/1.0.0/file',
            'schema_version': '1.0.0',
            'schema_type': 'file_bundle',
            'files': list(map(self.bundleFileIngest, files))
        }

    def bundle_protocols(self, protocols):
        return {
            'describedBy': 'https://schema.humancellatlas.org/bundle/5.1.0/protocol',
            'schema_type': 'protocol_bundle',
            'schema_version': '5.1.0',
            'protocols': list(map(self.bundleProtocol, protocols))
        }

    def bundle_links(self, links):
        return {
            'describedBy': 'https://schema.humancellatlas.org/bundle/1.0.0/links',
            'schema_type': 'link_bundle',
            'schema_version': '1.0.0',
            'links': links
        }

    def validate_metadata_files(self, bundle_contents):
        schema_version = {
            'project': None,
            'biomaterial': None,
            'process': None,
            'protocol': None,
            'file': '1.0.0',
            'links': '1.0.0'
        }

        for metadata_type, content in bundle_contents.items():
            # TODO fix bundle validator to only return the validation report
            result = self.bundle_validator.validate(content, metadata_type, schema_version[metadata_type])
            if not (True == result):
                raise InvalidBundleError('Invalid ' + metadata_type + ' bundle!')

    def create_bundle_manifest(self, submission_uuid, metadata_files_info, process_info):
        bundle_manifest = ingestapi.BundleManifest()

        bundle_manifest.envelopeUuid = submission_uuid

        bundle_manifest.dataFiles = list(process_info.derived_files.keys())

        bundle_manifest.fileProjectMap = {
            metadata_files_info['project']['dss_uuid']: list([process_info.project["uuid"]["uuid"]])
        }

        bundle_manifest.fileBiomaterialMap = {
            metadata_files_info['biomaterial']['dss_uuid']: list(process_info.input_biomaterials.keys())
        }

        bundle_manifest.fileProcessMap = {
            metadata_files_info['process']['dss_uuid']: list(process_info.derived_by_processes.keys())
        }

        bundle_manifest.fileProtocolMap = {
            metadata_files_info['protocol']['dss_uuid']: list(process_info.protocols.keys())
        }

        bundle_manifest.fileFilesMap = {
            metadata_files_info['file']['dss_uuid']: list(process_info.derived_files.keys())
        }

        # TODO store the version timestamp in bundle manifest
        # TODO do we need a fileLinksMap mapping in the bundle manifest?

        return bundle_manifest

    def dump_metadata_files_and_bundle_manifest(self, metadata_files_info, bundle_manifest):
        project = metadata_files_info['project']['content']
        project_keyword = project['content']['project_core']['project_shortname']

        for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
            bundle_file = metadata_files_info[metadata_type]
            self.dumpJsonToFile(bundle_file['content'], project_keyword, metadata_type + '_bundle')

        self.dumpJsonToFile(bundle_manifest.__dict__, project_keyword, 'bundleManifest')

    def upload_metadata_files(self, submission_uuid, metadata_files_info):
        try:
            for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
                bundle_file = metadata_files_info[metadata_type]
                filename = bundle_file['upload_filename']
                content = bundle_file['content']
                content_type = bundle_file['content_type']

                uploaded_file = self.writeMetadataToStaging(submission_uuid, filename, content, content_type)
                bundle_file['upload_file_url'] = uploaded_file.url
        except Exception as e:
            message = "An error occurred on uploading bundle files: " + str(e)
            raise BundleFileUploadError(message)

    def put_bundle_in_dss(self, bundle_uuid, created_files):
        try:
            created_bundle = self.dss_api.put_bundle(bundle_uuid, created_files)
        except Exception as e:
            message = 'An error occurred while putting bundle in DSS: ' + str(e)
            raise BundleDSSError(message)

        return created_bundle

    def put_files_in_dss(self, bundle_uuid, files_to_put):
        created_files = []

        for bundle_file in files_to_put:
            version = ''

            try:
                if bundle_file['is_same_as_input']:
                    # TODO this assumes that the latest version is the file version in the input bundle, should be a safe assumption for now
                    # Ideally, bundle manifest must store the file uuid and version and version must be retrieved from there
                    file_response = self.dss_api.head_file(bundle_file["dss_uuid"])
                    created_file = {
                        'version': file_response.headers['X-DSS-VERSION']
                    }
                else:
                    created_file = self.dss_api.put_file(bundle_uuid, bundle_file)


                version = created_file['version']
            except Exception as e:
                raise FileDSSError('An error occurred while putting file in DSS' + str(e))

            file_param = {
                "indexed": bundle_file["indexed"],
                "name": bundle_file["submittedName"],
                "uuid": bundle_file["dss_uuid"],
                "content-type": bundle_file["content-type"],
                "version": version
            }

            created_files.append(file_param)

        return created_files

    def get_metadata_files(self, metadata_files_info):
        metadata_files = []

        for metadata_type, metadata_file in metadata_files_info.items():
            metadata_files.append({
                'name': metadata_file['upload_filename'],
                'submittedName': metadata_file['dss_filename'],
                'url': metadata_file['upload_file_url'],
                'dss_uuid': metadata_file['dss_uuid'],
                'indexed': metadata_file['indexed'],
                'content-type': metadata_file['content_type'],
                'is_same_as_input': metadata_file['is_same_as_input'] if 'is_same_as_input' in metadata_file else False
            })

        return metadata_files

    def get_data_files(self, derived_files):
        data_files = []

        for uuid, data_file in derived_files.items():
            filename = data_file['fileName']
            cloud_url = data_file['cloudUrl']

            data_files.append({
                'name': filename,
                'submittedName': filename,
                'url': cloud_url,
                'dss_uuid': uuid,
                'indexed': False,
                'content-type': 'data',
                'is_same_as_input': data_file['is_same_as_input']
            })

        return data_files

    def prepare_data_files(self, process_info):
        for uuid, data_file in process_info.derived_files.items():
            process_info.derived_files[uuid]['is_same_as_input'] = uuid in process_info.input_files

        return process_info.derived_files

class File:
    def __init__(self):
        self.name = ""
        self.content_type = ""
        self.size = ""
        self.id = ""
        self.checksums = {}


class ProcessInfo:
    def __init__(self):
        self.project = {}

        # uuid => object mapping
        self.input_biomaterials = {}
        self.derived_by_processes = {}
        self.input_files = {}
        self.derived_files = {}
        self.protocols = {}

        self.links = []

        self.input_bundle = None


# Module Exceptions


class Error(Exception):
    """Base-class for all exceptions raised by this module."""


class MultipleProjectsError(Error):
    """A process should only have one project linked."""


class InvalidBundleError(Error):
    """There was a failure in bundle validation."""

class BundleFileUploadError(Error):
    """There was a failure in bundle file upload."""


class BundleDSSError(Error):
    """There was a failure in bundle creation in DSS."""


class FileDSSError(Error):
    """There was a failure in file creation in DSS."""


class NoUploadAreaFoundError(Error):
    """Export couldn't be as no upload area found"""

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    parser = OptionParser()
    parser.add_option("-e", "--subsEnvUuid", dest="submissionsEnvelopeUuid",
                      help="Submission envelope UUID for which to generate the bundle")
    parser.add_option("-p", "--processUrl", dest="processUrl",
                      help="Process Url")
    parser.add_option("-D", "--dry", help="do a dry run without submitting to ingest", action="store_true",
                      default=False)
    parser.add_option("-o", "--output", dest="output",
                      help="output directory where to dump json files submitted to ingest", metavar="FILE",
                      default=None)
    parser.add_option("-i", "--ingest", help="the URL to the ingest API")
    parser.add_option("-s", "--staging", help="the URL to the staging API")
    parser.add_option("-d", "--dss", help="the URL to the datastore service")
    parser.add_option("-l", "--log", help="the logging level", default='INFO')
    parser.add_option("-v", "--version", dest="schema_version", help="Metadata schema version", default=None)

    (options, args) = parser.parse_args()

    if not options.submissionsEnvelopeUuid:
        print ("You must supply a submission envelope UUID")
        exit(2)

    if not options.processUrl:
        print ("You must supply a processUrl")
        exit(2)

    ex = IngestExporter(options)

    ex.export_bundle(options.submissionsEnvelopeUuid, options.processUrl)



#!/usr/bin/env python
"""
Description goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "12/09/2017"

import json
import os
import requests
import logging

from time import sleep
from urllib.parse import urlparse

DEFAULT_STAGING_URL = os.environ.get('STAGING_API', 'https://staging.dev.data.humancellatlas.org')
DEFAULT_STAGING_VERSION = os.environ.get('STAGING_API_VERSION', 'v1')
INGEST_API_KEY = os.environ.get('INGEST_API_KEY', 'zero-pupil-until-funny')


class StagingApi:
    def __init__(self, url=None, apikey=None, apiversion=None):
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        logging.basicConfig(formatter=formatter)
        logging.getLogger("requests").setLevel(logging.WARNING)
        self.logger = logging.getLogger(__name__)

        if not apiversion and DEFAULT_STAGING_VERSION:
            self.apiversion = DEFAULT_STAGING_VERSION
        self.apiversion = apiversion if apiversion else "v1"

        if not url and DEFAULT_STAGING_URL:
            url = DEFAULT_STAGING_URL
            # expand interpolated env vars
            self.url = os.path.expandvars(url)
            self.logger.info("using " + url + " for staging API")
        self.url = url if url else "https://staging.dev.data.humancellatlas.org"

        if not apikey and INGEST_API_KEY:
            apikey = INGEST_API_KEY
        self.apikey = apikey if apikey else "zero-pupil-until-funny"

        self.header = {'Api-Key': self.apikey, 'Content-type': 'application/json'}

    def createStagingArea(self, submissionId):
        base = urlparse.urljoin(self.url, self.apiversion + '/area/' + submissionId)
        r = requests.post(base, headers=self.header)
        if r.status_code == requests.codes.created:
            print ("Waiting 10 seconds for IAM policy to take effect..."),
            sleep(10)
            print ("staging area created!:" + base)
            return json.loads(r.text)

        raise ValueError('Can\'t create staging area for sub id:' + submissionId + ', Error:' + r.text)

    def deleteStagingArea(self, submissionId):
        base = urlparse.urljoin(self.url, self.apiversion + '/area/' + submissionId)
        try:

            r = requests.delete(base, headers=self.header)
            if r.status_code == requests.codes.no_content:
                print ("staging area deleted!")
                return base
            else:
                return base
        except:
            raise ValueError('Can\'t delete staging area for sub id:' + submissionId)

    def stageFile(self, submissionId, filename, body, type):

        fileUrl = urlparse.urljoin(self.url, self.apiversion + '/area/' + submissionId + "/" + filename)

        header = dict(self.header)
        header['Content-type'] = 'application/json; dcp-type=' + type

        r = self._retry_when_http_error(0, self._put_file, fileUrl, body, header)

        if r and (r.status_code == requests.codes.ok or requests.codes.created):
            responseObject = json.loads(r.text)
            return FileDescription(responseObject["checksums"], type, responseObject["name"], responseObject["size"],
                                   responseObject["url"], )

        raise ValueError('Can\'t create staging area for sub id:' + submissionId)

    def _put_file(self, fileUrl, body, header):
        r = requests.put(fileUrl, data=json.dumps(body, indent=4), headers=header)
        return r

    def hasStagingArea(self, submissionId):
        base = urlparse.urljoin(self.url, self.apiversion + '/area/' + submissionId + '/files_info')

        r = requests.put(base, data="{}", headers=self.header)
        return r.status_code == requests.codes.ok

    """
        func should return http response r
    """

    def _retry_when_http_error(self, tries, func, *args):
        max_retries = 5

        if tries < max_retries:
            self.logger.debug("no of tries: " + str(tries + 1))
            r = func(*args)

            try:
                r.raise_for_status()
            except requests.HTTPError:
                self.logger.error("\nResponse was: " + str(r.status_code) + " (" + r.text + ")")
                tries += 1
                sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            return r
        else:
            error_message = "Maximum no of tries reached: " + str(max_retries)
            self.logger.error(error_message)
            return None


class FileDescription:
    def __init__(self, checksums, contentType, name, size, url):
        self.checksums = checksums
        self.content_type = contentType
        self.name = name
        self.size = size
        self.url = url



#!/usr/bin/env python
"""
desc goes here
"""
import time
from requests import HTTPError

__author__ = "jupp"
__license__ = "Apache 2.0"

import json, os, requests, logging, uuid


class IngestApi:
    def __init__(self, url=None):
        formatter = logging.Formatter(
            '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s')
        logging.basicConfig(formatter=formatter)
        logging.getLogger("requests").setLevel(logging.WARNING)
        self.logger = logging.getLogger(__name__)

        if not url and 'INGEST_API' in os.environ:
            url = os.environ['INGEST_API']
            # expand interpolated env vars
            url = os.path.expandvars(url)
            self.logger.info("using " + url + " for ingest API")
        self.url = url if url else "http://localhost:8080"

        self.ingest_api = None
        self.headers = {'Content-type': 'application/json'}

        self.submission_links = {}
        self.load_root()

    def load_root(self):
        if not self.ingest_api:
            reply = requests.get(self.url, headers=self.headers)
            self.ingest_api = reply.json()["_links"]

    def getSubmissions(self):
        params = {'sort': 'submissionDate,desc'}
        r = requests.get(self.ingest_api["submissionEnvelopes"]["href"].rsplit("{")[0], params=params,
                         headers=self.headers)
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["_embedded"]["submissionEnvelopes"]

    def getSubmissionIfModifiedSince(self, submissionId, datetimeUTC):
        submissionUrl = self.getSubmissionUri(submissionId)
        headers = self.headers

        if datetimeUTC:
            headers = {'If-Modified-Since': datetimeUTC}

        self.logger.info('headers:' + str(headers))
        r = requests.get(submissionUrl, headers=headers)

        if r.status_code == requests.codes.ok:
            submission = json.loads(r.text)
            return submission
        else:
            self.logger.error(str(r))

    def getProjects(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/projects'
        r = requests.get(submissionUrl, headers=self.headers)
        projects = []
        if r.status_code == requests.codes.ok:
            projects = json.loads(r.text)
        return projects

    def getProjectById(self, id):
        submissionUrl = self.url + '/projects/' + id
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            project = json.loads(r.text)
            return project
        else:
            raise ValueError("Project " + id + " could not be retrieved")

    def getProjectByUuid(self, uuid):
        url =  self.url + '/projects/search/findByUuid?uuid=' + uuid
        r = requests.get(url, headers=self.headers)
        r.raise_for_status()
        return r.json()

    def getSubmissionEnvelope(self, submissionUrl):
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            submissionEnvelope = json.loads(r.text)
            return submissionEnvelope
        else:
            raise ValueError("Submission Envelope " + submissionUrl + " could not be retrieved")

    def getFiles(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/files'
        r = requests.get(submissionUrl, headers=self.headers)
        files = []
        if r.status_code == requests.codes.ok:
            files = json.loads(r.text)
        return files

    def getBundleManifests(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/bundleManifests'
        r = requests.get(submissionUrl, headers=self.headers)
        bundleManifests = []

        if r.status_code == requests.codes.ok:
            bundleManifests = json.loads(r.text)
        return bundleManifests

    def createSubmission(self, token):
        auth_headers = {'Content-type': 'application/json',
                        'Authorization': token
                        }
        try:
            r = requests.post(self.ingest_api["submissionEnvelopes"]["href"].rsplit("{")[0], data="{}",
                              headers=auth_headers)
            r.raise_for_status()
            submissionUrl = json.loads(r.text)["_links"]["self"]["href"].rsplit("{")[0]
            self.submission_links[submissionUrl] = json.loads(r.text)["_links"]
            return submissionUrl
        except requests.exceptions.RequestException as err:
            self.logger.error("Request failed: ", err)
            raise

    def finishSubmission(self, submissionUrl):
        r = requests.put(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.update:
            self.logger.info("Submission complete!")
            return r.text

    def updateSubmissionState(self, submissionId, state):
        state_url = self.getSubmissionStateUrl(submissionId, state)

        if state_url:
            r = requests.put(state_url, headers=self.headers)

        return self.handleResponse(r)

    def getSubmissionStateUrl(self, submissionId, state):
        submissionUrl = self.getSubmissionUri(submissionId)
        response = requests.get(submissionUrl, headers=self.headers)
        submission = self.handleResponse(response)

        if submission and state in submission['_links']:
            return submission['_links'][state]["href"].rsplit("{")[0]

        return None

    def handleResponse(self, response):
        if response.ok:
            return json.loads(response.text)
        else:
            self.logger.error('Response:' + response.text)
            return None

    def getSubmissionUri(self, submissionId):
        return self.ingest_api["submissionEnvelopes"]["href"].rsplit("{")[0] + "/" + submissionId

    def getAssayUrl(self, assayCallbackLink):
        # TODO check if callback link already has a leading slash
        return self.url + "/" + assayCallbackLink

    def getAssay(self, assayUrl):
        r = requests.get(assayUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            return r.json()


    def getAnalyses(self, submissionUrl):
        return self.getEntities(submissionUrl, "analyses")

    def getEntities(self, submissionUrl, entityType):
        r = requests.get(submissionUrl, headers=self.headers)
        if r.status_code == requests.codes.ok:
            if entityType in json.loads(r.text)["_links"]:
                # r2 = requests.get(, headers=self.headers)
                for entity in self._getAllObjectsFromSet(json.loads(r.text)["_links"][entityType]["href"], entityType):
                    yield entity

    def _getAllObjectsFromSet(self, url, entityType, pageSize=None):
        params = dict()
        if pageSize:
            params = {"size": pageSize}

        r = requests.get(url, headers=self.headers, params=params)
        if r.status_code == requests.codes.ok:
            if "_embedded" in json.loads(r.text):
                for entity in json.loads(r.text)["_embedded"][entityType]:
                    yield entity
                if "next" in json.loads(r.text)["_links"]:
                    for entity2 in self._getAllObjectsFromSet(json.loads(r.text)["_links"]["next"]["href"], entityType):
                        yield entity2

    def getRelatedEntities(self, relation, entity, entityType):
        # get the self link from entity
        if relation in entity["_links"]:
            entityUri = entity["_links"][relation]["href"]
            for entity in self._getAllObjectsFromSet(entityUri, entityType):
                yield entity

    def _updateStatusToPending(self, submissionUrl):
        r = requests.patch(submissionUrl, data="{\"submissionStatus\" : \"Pending\"}", headers=self.headers)

    def createProject(self, submissionUrl, jsonObject, token):
        return self.createEntity(submissionUrl, jsonObject, "projects", token)

    def createBiomaterial(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "biomaterials")

    def createProcess(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "processes")

    # def createDonor(self, submissionUrl, jsonObject):
    #     return self.createBiomaterial(submissionUrl, jsonObject)

    def createProtocol(self, submissionUrl, jsonObject):
        return self.createEntity(submissionUrl, jsonObject, "protocols")

    # def createAnalysis(self, submissionUrl, jsonObject):
    #     return self.createEntity(submissionUrl, jsonObject, "analyses")

    def createFile(self, submissionUrl, fileName, jsonObject):
        submissionUrl = self.submission_links[submissionUrl]["files"]['href'].rsplit("{")[0]
        self.logger.debug("posting " + submissionUrl)
        fileToCreateObject = {
            "fileName": fileName,
            "content": json.loads(jsonObject)
        }
        r = requests.post(submissionUrl, data=json.dumps(fileToCreateObject),
                          headers=self.headers)
        if r.status_code == requests.codes.created or r.status_code == requests.codes.accepted:
            return json.loads(r.text)
        raise ValueError('Create file failed: File ' + fileName + " - " + r.text)

    def createEntity(self, submissionUrl, jsonObject, entityType, token=None):
        self.logger.debug(".", )
        auth_headers = {'Content-type': 'application/json',
                        'Authorization': token
                        }
        submissionUrl = self.submission_links[submissionUrl][entityType]['href'].rsplit("{")[0]

        self.logger.debug("posting " + submissionUrl)
        r = requests.post(submissionUrl, data=jsonObject,
                          headers=auth_headers)
        if r.status_code == requests.codes.created or r.status_code == requests.codes.accepted:
            return json.loads(r.text)

    # given a HCA object return the URI for the object from ingest
    def getObjectId(self, entity):
        if "_links" in entity:
            entityUrl = entity["_links"]["self"]["href"].rsplit("{")[0]
            return entityUrl
        raise ValueError('Can\'t get id for ' + json.dumps(entity) + ' is it a HCA entity?')

    def getObjectUuid(self, entityUri):
        r = requests.get(entityUri,
                         headers=self.headers)
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["uuid"]["uuid"]

    def linkEntity(self, fromEntity, toEntity, relationship):
        if not fromEntity:
            raise ValueError("Error: fromEntity is None")

        if not toEntity:
            raise ValueError("Error: toEntity is None")

        if not relationship:
            raise ValueError("Error: relationship is None")

        # check each dict in turn for non-None-ness

        fromEntityLinks = fromEntity["_links"] if "_links" in fromEntity else None
        if not fromEntityLinks:
            raise ValueError("Error: fromEntity has no _links")

        fromEntityLinksRelationship = fromEntityLinks[relationship] if relationship in fromEntityLinks else None
        if not fromEntityLinksRelationship:
            raise ValueError("Error: fromEntityLinks has no {0} relationship".format(relationship))

        fromEntityLinksRelationshipHref = fromEntityLinksRelationship["href"] if "href" in fromEntityLinksRelationship else None
        if not fromEntityLinksRelationshipHref:
            raise ValueError("Error: fromEntityLinksRelationship for relationship {0} has no href".format(relationship))

        fromUri = fromEntity["_links"][relationship]["href"]
        toUri = self.getObjectId(toEntity)

        self._retry_when_http_error(0, self._post_link_entity, fromUri, toUri)

    def _post_link_entity(self, fromUri, toUri):
        self.logger.debug('fromUri ' + fromUri + ' toUri:' + toUri);

        headers = {'Content-type': 'text/uri-list'}

        r = requests.post(fromUri.rsplit("{")[0],
                          data=toUri.rsplit("{")[0], headers=headers)

        return r

    def _retry_when_http_error(self, tries, func, *args):
        max_retries = 5

        if tries < max_retries:
            self.logger.info("no of tries: " + str(tries + 1))
            r = None
            
            try:
                r = func(*args)
                r.raise_for_status()

            except HTTPError:
                self.logger.error("\nResponse was: " + str(r.status_code) + " (" + r.text + ")")
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            except requests.ConnectionError as e:
                self.logger.exception(str(e))
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            except Exception as e:
                self.logger.exception(str(e))
                tries += 1
                time.sleep(1)
                r = self._retry_when_http_error(tries, func, *args)

            return r
        else:
            error_message = "Maximum no of tries reached: " + str(max_retries)
            self.logger.error(error_message)
            return None

    def _request_post(self, url, data, params, headers):
        if params:
            return requests.post(url, data=data, params=params, headers=headers)

        return requests.post(url, data=data, headers=headers)

    def _request_put(self, url, data, params, headers):
        if params:
            return requests.put(url, data=data, params=params, headers=headers)

        return requests.put(url, data=data, headers=headers)

    def createBundleManifest(self, bundleManifest):
        r = self._retry_when_http_error(0, self._post_bundle_manifest, bundleManifest, self.ingest_api["bundleManifests"]["href"].rsplit("{")[0])

        if not (200 <= r.status_code < 300):
            error_message = "Failed to create bundle manifest at URL {0} with request payload: {1}".format(self.ingest_api["bundleManifests"]["href"].rsplit("{")[0],
                                                                                                           json.dumps(bundleManifest.__dict__))
            self.logger.error(error_message)
            raise ValueError(error_message)
        else:
            self.logger.info("successfully created bundle manifest")

    def _post_bundle_manifest(self, bundleManifest, url):
        return requests.post(url, data=json.dumps(bundleManifest.__dict__), headers=self.headers)

    def updateSubmissionWithStagingCredentials(self, subUrl, uuid, submissionCredentials):
        stagingDetails = \
            {
                "stagingDetails": {
                    "stagingAreaUuid": {
                        "uuid": uuid
                    },
                    "stagingAreaLocation": {
                        "value": submissionCredentials
                    }
                }
            }

        if self.retrySubmissionUpdateWithStagingDetails(subUrl, stagingDetails, 0):
            self.logger.debug("envelope updated with staging details " + json.dumps(stagingDetails))
        else:
            self.logger.error("Failed to update envelope with staging details: " + json.dumps(stagingDetails))

    def retrySubmissionUpdateWithStagingDetails(self, subUrl, stagingDetails, tries):
        if tries < 5:
            # do a GET request to get latest submission envelope
            entity_response = requests.get(subUrl)
            etag = entity_response.headers['ETag']
            if etag:
                # set the etag header so we get 412 if someone beats us to set validating
                self.headers['If-Match'] = etag
                r = requests.patch(subUrl, data=json.dumps(stagingDetails))
                try:
                    r.raise_for_status()
                    return True
                except HTTPError:
                    self.logger.error("PATCHing submission envelope with creds failed, retrying")
                    tries += 1
                    self.retrySubmissionUpdateWithStagingDetails(subUrl, stagingDetails, tries)
        else:
            return False


class BundleManifest:
    def __init__(self):
        self.bundleUuid = str(uuid.uuid4())
        self.envelopeUuid = {}
        self.dataFiles = []
        self.fileBiomaterialMap = {}
        self.fileProcessMap = {}
        self.fileFilesMap = {}
        self.fileProjectMap = {}
        self.fileProtocolMap = {}



#!/usr/bin/env python
"""
Description goes here
"""
import datetime
import hca
import json
import logging
import os


__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "12/09/2017"


class DssApi:
    def __init__(self, url=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        logging.getLogger("requests").setLevel(logging.WARNING)
        self.logger = logging.getLogger(__name__)

        self.url = url if url else "http://dss.dev.data.humancellatlas.org"

        if not url and 'DSS_API' in os.environ:
            url = os.environ['DSS_API']
            # expand interpolated env vars
            self.url = os.path.expandvars(url)
            self.logger.info("using " + url + " for dss API")

        self.headers = {'Content-type': 'application/json'}

        self.hca_client = hca.dss.DSSClient()
        self.hca_client.host = self.url + "/v1"
        self.creator_uid = 8008

    def put_file(self, bundle_uuid, file):
        url = file["url"]
        uuid = file["dss_uuid"]

        try:
            bundle_file = self.hca_client.put_file(
                uuid=uuid,
                bundle_uuid=bundle_uuid,
                creator_uid=self.creator_uid,
                source_url=url
            )
        except Exception as e:
            params = {
                'uuid': uuid,
                'bundle_uuid': bundle_uuid,
                'creator_uid': self.creator_uid,
                'source_url': url
            }
            self.logger.error('Error in hca_client.put_file method call with params:' + json.dumps(params))
            raise Error(e)

        return bundle_file

    def put_bundle(self, bundle_uuid, bundle_files):
        bundle = None

        # Generate version client-side for idempotent PUT /bundle
        version = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S.%fZ")

        # finally create the bundle
        try:
            bundle = self.hca_client.put_bundle(
                uuid=bundle_uuid,
                version=version,
                replica="aws",
                files=bundle_files,
                creator_uid=self.creator_uid
            )
        except Exception as e:
            params = {
                'uuid': bundle_uuid,
                'version': version,
                'replica': "aws",
                'files': bundle_files,
                'creator_uid': self.creator_uid
            }
            self.logger.error('Error in hca_client.put_bundle method call with params:' + json.dumps(params))
            raise Error(e)

        return bundle

    def head_file(self, file_uuid ):
        # finally create the bundle
        try:
            return self.hca_client.head_file(
                uuid=file_uuid,
                replica="aws"
            )
        except Exception as e:
            raise Error(e)


# Module Exceptions


class Error(Exception):
    """Base-class for all exceptions raised by this module."""
